title: Web Components as Progressive Enhancement 
url: https://cloudfour.com/thinks/web-components-as-progressive-enhancement/
hash_url: 72b4d15d286c2df85aba5742b1d98e62

<div class="u-bgGray u-pad1 u-pullSides1 u-spaceItems1 u-textGrow1"><p>Web components are a powerful tool for creating custom HTML elements, but you can run into challenges if you try to replace existing elements with web components. To get the best of both worlds, try wrapping existing elements in web components instead.</p></div>
<p>I learned this the hard way&#8230; On a few recent projects we&#8217;ve wanted to create an auto-expanding <code>textarea</code>: When a user types its height would increase so that its content is never clipped.</p>
<p>I was tired of rewriting this behavior across projects and frameworks and began thinking through how it could be written as a reusable web component. My first draft of an API looked like this:</p>
<pre><code class="language-html">&lt;elastic-textarea name="textarea-name" id="textarea-id" rows="4"&gt;
  Here is my textarea content
&lt;/elastic-textarea&gt;
</code></pre>
<p>This is pretty sweet. Developers can use this exactly like the <code>textarea</code> they&#8217;re already used to! But, as I was testing this out I noticed a couple of major downsides:</p>
<ol>
<li>Web components require JavaScript to render <sup id="fnref-6408-1"><a href="#fn-6408-1" class="jetpack-footnote" title="Read footnote.">1</a></sup>. This means that until the JavaScript is downloaded, parsed, and run, no <code>textarea</code> is displayed. Instead &#8220;Here is my textarea content&#8221; is displayed as an unstyled string. If a user has JavaScript disabled, or the JavaScript fails to load, the textarea just won&#8217;t work.</li>
<li>The <code>textarea</code> element has its own complex behavior, APIs, and accessibility behaviors. I&#8217;d need to recreate a lot of this from scratch for my component. I&#8217;d need to apply the <code>name</code>, <code>id</code>, and <code>rows</code> to the actual <code>textarea</code> element that my JavaScript renders. I&#8217;d also need to add all of the <code>textarea</code> JavaScript APIs to my custom element (<code>.value</code>, <code>.is-valid</code>, etc.) <sup id="fnref-6408-2"><a href="#fn-6408-2" class="jetpack-footnote" title="Read footnote.">2</a></sup></li>
</ol>
<div class="HashHeading HashHeading--h2">
  <h2 id="the-solution">The Solution</h2>
  <a href="#the-solution"
    aria-label="Permalink for The Solution"
    class="HashHeading-link">
    <svg viewBox="0 0 24 24" width="24" height="24" class="Icon" role="presentation"><g fill="none" stroke="currentColor" stroke-width="3" stroke-linecap="round"><path d="M13.77,10.09l-0.71-.71a4,4,0,0,0-5.65,0L3.16,13.63a4,4,0,0,0,0,5.66l1.4,1.4a4,4,0,0,0,5.67,0l1.41-1.41"/><path d="M10.23,13.62l0.71,0.71a4,4,0,0,0,5.65,0l4.25-4.25a4,4,0,0,0,0-5.66L19.43,3a4,4,0,0,0-5.67,0L12.35,4.43"/></g></svg>

  </a>
</div>
  
<p>We can avoid both of these drawbacks by wrapping and enhancing an existing element instead of replacing it. Here&#8217;s what that looks like:</p>
<pre><code class="language-html">&lt;elastic-textarea&gt;
  &lt;textarea name="textarea-name" id="textarea-id" rows="4"&gt;
    Here is my textarea content
  &lt;/textarea&gt;
&lt;/elastic-textarea&gt;
</code></pre>
<p>This is a little more verbose, but it gets me the best of both worlds. Before JavaScript loads I have a fully functioning <code>textarea</code>. After JavaScript loads my <code>textarea</code> is progressively enhanced with additional functionality.</p>
<p>It could get a little tedious to wrap every <code>textarea</code> on the page. To avoid this we could set the component up so it can wrap multiple <code>textarea</code>s and enhance all of them at once. This would allow you to wrap a whole form or page to enable this behavior:</p>
<pre><code class="language-html">&lt;elastic-textarea&gt;
  &lt;textarea name="textarea-name" id="textarea-id" rows="2"&gt;
    Here is my textarea content
  &lt;/textarea&gt;

  &lt;textarea name="textarea-2" id="textarea-id-2" rows="5"&gt;
    Here is another textarea
  &lt;/textarea&gt;
&lt;/elastic-textarea&gt;
</code></pre>
<p>You can play with the finished component below:</p>
<figure class="Figure" style="text-align: center;">
  <script type="module" src="https://unpkg.com/@cloudfour/elastic-textarea/index.min.js"></script><br />
  <elastic-textarea><br />
    <label for="textarea-id">Type in the textarea below to watch it expand and contract.</label><br />
    <textarea name="textarea-name" id="textarea-id" style="resize: horizontal;width:100%;margin-top: 1em;"></textarea><br />
  </elastic-textarea><figcaption class="Figure-caption">
    We released this as an <a href="https://www.npmjs.com/package/@cloudfour/elastic-textarea">open source component</a> you can use in your projects. Most of the JavaScript logic was borrowed from an excellent implementation my colleague <a href="https://cloudfour.com/is/scott">Scott Vandehey</a> wrote for a previous project.<br />
  </figcaption></figure>
<div class="HashHeading HashHeading--h2">
  <h2 id="the-skys-the-limit">The Sky&#8217;s the Limit</h2>
  <a href="#the-skys-the-limit"
    aria-label="Permalink for The Sky&#8217;s the Limit"
    class="HashHeading-link">
    <svg viewBox="0 0 24 24" width="24" height="24" class="Icon" role="presentation"><g fill="none" stroke="currentColor" stroke-width="3" stroke-linecap="round"><path d="M13.77,10.09l-0.71-.71a4,4,0,0,0-5.65,0L3.16,13.63a4,4,0,0,0,0,5.66l1.4,1.4a4,4,0,0,0,5.67,0l1.41-1.41"/><path d="M10.23,13.62l0.71,0.71a4,4,0,0,0,5.65,0l4.25-4.25a4,4,0,0,0,0-5.66L19.43,3a4,4,0,0,0-5.67,0L12.35,4.43"/></g></svg>

  </a>
</div>
  
<p>The <code>elastic-textarea</code> component is a single example of how web components can be used for progressive enhancement, but there are tons of other potential use cases for this technique. Here are a couple other examples to check out for inspiration:</p>
<ul>
<li><a href="https://www.zachleat.com/">Zach Leatherman</a> created an <em>excellent</em> <a href="https://www.zachleat.com/web/details-utils/"><code>details-utils</code></a> component that wraps and progressively enhances the <a href="https://developer.mozilla.org/en-US/docs/Web/HTML/Element/details"><code>details</code> element</a>. (This was a big source of inspiration for my approach here.)</li>
<li>I used this same technique to create <a href="https://cloudfour.com/thinks/building-an-accessible-image-comparison-web-component/">a progressively enhanced image comparison component</a>.</li>
</ul>
<div class="HashHeading HashHeading--h2">
  <h2 id="sharing-is-caring">Sharing is Caring</h2>
  <a href="#sharing-is-caring"
    aria-label="Permalink for Sharing is Caring"
    class="HashHeading-link">
    <svg viewBox="0 0 24 24" width="24" height="24" class="Icon" role="presentation"><g fill="none" stroke="currentColor" stroke-width="3" stroke-linecap="round"><path d="M13.77,10.09l-0.71-.71a4,4,0,0,0-5.65,0L3.16,13.63a4,4,0,0,0,0,5.66l1.4,1.4a4,4,0,0,0,5.67,0l1.41-1.41"/><path d="M10.23,13.62l0.71,0.71a4,4,0,0,0,5.65,0l4.25-4.25a4,4,0,0,0,0-5.66L19.43,3a4,4,0,0,0-5.67,0L12.35,4.43"/></g></svg>

  </a>
</div>
  
<p>I&#8217;m a big fan of this strategy and I&#8217;m excited to create and share more progressively enhanced components. At Cloud Four we work on a lot of different projects for a lot of different clients, and I&#8217;ve found myself rewriting the same functionality across several projects. Next time I catch myself rewriting the same component logic for the third or fourth time I&#8217;m going to make it a web component so I don&#8217;t have to write it a fifth time.</p>
<p>By packaging these chunks of interactive logic as custom elements, we can make them easy to share and reuse across projects and frameworks. By enhancing native HTML instead of replacing it, we can provide a solid baseline experience, and add progressive enhancement as the cherry on top.</p>
<div class="footnotes">
<hr />
<ol>
<li id="fn-6408-1">
There are various libraries that allow server side rendering of web components, but they usually require you to buy into a specific web component framework, and/or modify your build steps. These are useful tools for individuals projects, but don&#8217;t work as well for creating easily shareable components.&#160;<a href="#fnref-6408-1" title="Return to main content."><svg viewBox="0 0 24 24" width="24" height="24" class="Icon" role="presentation"><g fill="none" stroke="currentColor" stroke-width="4" stroke-linecap="round" stroke-linejoin="round"><path d="M4,14H16a5.76,5.76,0,0,0,6-6,5.76,5.76,0,0,0-6-6"/><polyline points="8 8 2 14 8 20"/></g></svg>
</a>
</li>
<li id="fn-6408-2">
There&#8217;s a spec to allow <a href="https://web.dev/custom-elements-v1/#extending-native-html-elements">extending existing elements</a> but it looks unlikely that Safari will adopt it. On some projects, <a href="https://cloudfour.com/thinks/mighty-morphin-web-components/">recreating element APIs can have a big payoff</a> but it&#8217;s a lot of work if you&#8217;re just trying to add a little functionality to an existing element.&#160;<a href="#fnref-6408-2" title="Return to main content."><svg viewBox="0 0 24 24" width="24" height="24" class="Icon" role="presentation"><g fill="none" stroke="currentColor" stroke-width="4" stroke-linecap="round" stroke-linejoin="round"><path d="M4,14H16a5.76,5.76,0,0,0,6-6,5.76,5.76,0,0,0-6-6"/><polyline points="8 8 2 14 8 20"/></g></svg>
</a>
</li>
</ol>
</div>
