title: Better Buttons with color-mix() and Custom Properties
url: https://www.abeautifulsite.net/posts/better-buttons-with-color-mix-and-custom-properties/
hash_url: b2ccbe813d97ba39ba64a1b6d14422dd
archive_date: 2024-02-28
og_image: http://0.gravatar.com/avatar/bf1b3b95fd5b096a3592247c29667b33?s=512
description: Let the browser and OKLCH do the hard work of styling states.
favicon: https://www.abeautifulsite.net/images/logos/leaf.svg
language: en_US

<p>Let's build a button that accepts one color and calculates its hover and focus states automatically. For this experiment, we'll use <a href="https://developer.mozilla.org/en-US/docs/Web/CSS/Using_CSS_custom_properties">CSS Custom Properties</a>, <a href="https://developer.mozilla.org/en-US/docs/Web/CSS/color_value/color-mix">color-mix()</a>, and <a href="https://developer.mozilla.org/en-US/docs/Web/CSS/color_value/oklch">OKLCH</a> to ensure that tints and shades are perceptually uniform.</p>
<p>To keep things simple, we'll follow today's flat design trend and use a single color for the background. We'll store that color in a custom property so we can reuse it and change things up as we go.</p>
<p>Let's kick things off with some styles.</p>
<pre><code class="language-css">/* Set some defaults for the button */
:root {
  --button-bg: slategrey;
}

/* Remove factory button styles and add our own */
button {
  unset: all;
  font: 16px sans-serif;
  border: none;
  border-radius: 6px;
  background-color: var(--button-bg);
  color: white;
  padding: .875rem 1.125rem;
  margin: 0;
  cursor: pointer;
  transition: 100ms background-color;
}

/* Tint the background when hovering */
button:hover {
  background-color: color-mix(in oklch, var(--button-bg) 100%, white 8%);
}

/* Shade the background when clicked */
button:active {
  background-color: color-mix(in oklch, var(--button-bg) 100%, black 4%);
}

/* Remove the default focus ring */
button:focus {
  outline: none;
}

/* Provide a color-matched focus ring for keyboard users */
button:focus-visible {
  outline: solid 3px var(--button-bg);
  outline-offset: 1px;
}
</code></pre>
<p>We're using a custom property called <code>--button-bg</code> that sets the button's background. Using a custom property allows us to pass it into the <code>color-mix()</code> function to calculate the hover and active states. This also allows us to use it again to make a color-matched outline for the focus state.</p>
<p>The magic is here:</p>
<pre><code class="language-css">/* Tint the background when hovering */
button:hover {
  background-color: color-mix(in oklch, var(--button-bg) 100%, white 8%);
}

/* Shade the background when clicked */
button:active {
  background-color: color-mix(in oklch, var(--button-bg) 100%, black 4%);
}
</code></pre>
<p>Using <code>color-mix()</code>, we can adjust the tint/shade based on the background color, meaning we don't need to manually select lighter/darker colors for those states. And because we're using OKLCH, the variations will be perceptually uniform, <a href="https://lea.verou.me/blog/2020/04/lch-colors-in-css-what-why-and-how/#3.-lch-lightness-actually-means-something">unlike HSL</a>. This means that tints and shades will look consistent for any color we choose!</p>
<p>Oh, and because we're using a custom property, we can easily test that theory by setting <code>--button-bg</code> to any color. Note how the hover and focus states update accordingly in this example.*</p>

<p>The caveat is <a href="https://caniuse.com/?search=color-mix">browser support for <code>color-mix()</code></a>, which is green across the board, but still somewhat recent. However, I would argue that this is a great candidate for <a href="https://developer.mozilla.org/en-US/docs/Glossary/Progressive_Enhancement">progressive enhancement</a>, as the absence of color changes on hover and click doesn't affect the function of the button.</p>
<p>Not bad for a few lines of CSS!</p>
<hr>
<small>
</small><p>*Alas, we can't automatically ensure the label's contrast based on the background color, but when <a href="https://www.w3.org/TR/css-color-5/">CSS Color Module Level 5</a> lands, we'll get <a href="https://caniuse.com/css-relative-colors#:~:text=The%20CSS%20Relative%20Color%20syntax,any%20of%20the%20color%20values.">relative colors</a> that will <a href="https://twitter.com/anatudor/status/1693855796499468720">allow us to do just that</a>!</p>
