title: i18n with tagged template strings in ECMAScript 6
url: http://jaysoo.ca/2014/03/20/i18n-with-es6-template-strings/
hash_url: 68f2d862066c644a50f929e8a787406b

<p>One of the new features coming to ECMAScript 6 (ES6), the next version of
JavaScript standards, is the <strong>template string</strong>. The simplest use cases for
template strings are creating multiline strings, and doing string interpolation.</p>

<p>Multiline string:</p>

<div class="highlight"><pre><code class="javascript"><span class="kd">let</span> <span class="nx">s</span> <span class="o">=</span> <span class="err">`</span><span class="nx">This</span> <span class="nx">is</span>
<span class="nx">a</span> <span class="nx">multiline</span>
<span class="nx">string</span><span class="err">`</span><span class="p">;</span>
</code></pre></div>

<p>String interpolation:</p>

<div class="highlight"><pre><code class="javascript"><span class="kd">let</span> <span class="nx">firstName</span> <span class="o">=</span> <span class="s1">'Bob'</span><span class="p">,</span> <span class="nx">lastName</span> <span class="o">=</span> <span class="s1">'Smith'</span><span class="p">;</span>
<span class="kd">let</span> <span class="nx">msg</span> <span class="o">=</span> <span class="err">`</span><span class="nx">Hello</span> <span class="nx">$</span><span class="p">{</span><span class="nx">firstName</span><span class="p">}</span> <span class="nx">$</span><span class="p">{</span><span class="nx">lastName</span><span class="p">}</span><span class="o">!</span><span class="err">`</span><span class="p">;</span> <span class="c1">// 'Hello Bob Smith!'</span>
</code></pre></div>

<p>You can also <strong>tag</strong> a template string by adding an expression before it.
When a template string is tagged, the literals and substitutions are passed to
the tag function, and whatever returns from the function is the resulting value.</p>

<div class="highlight"><pre><code class="javascript"><span class="kd">function</span> <span class="nx">foo</span><span class="p">(</span><span class="nx">literals</span><span class="p">,</span> <span class="p">...</span><span class="nx">values</span><span class="p">)</span> <span class="p">{</span>
  <span class="k">return</span> <span class="mi">42</span><span class="p">;</span>
<span class="p">}</span>

<span class="nx">foo</span><span class="err">`</span><span class="nx">What</span> <span class="k">do</span> <span class="nx">you</span> <span class="nx">get</span> <span class="k">if</span> <span class="nx">you</span> <span class="nx">multiply</span> <span class="nx">$</span><span class="p">{</span><span class="mi">6</span><span class="p">}</span> <span class="nx">by</span> <span class="nx">$</span><span class="p">{</span><span class="mi">9</span><span class="p">}</span><span class="o">?</span><span class="err">`</span><span class="p">;</span> <span class="c1">// 42</span>

<span class="c1">// In the foo function:</span>
<span class="c1">//   literals = ['What do you  get if you multiply ', ' by ', '?']</span>
<span class="c1">//   values[0] = 6</span>
<span class="c1">//   values[1] = 9</span>
</code></pre></div>

<p>This feature opens up some interesting syntactic sugar for producing and manipulating
content. One such example is <strong>Internationalization</strong> or <strong>i18n</strong>.</p>

<p>In this post weâll look at:</p>

<ul>
  <li>Implementing a basic i18n tag function for translating messages.</li>
  <li>Localization (l10n) of values using the <strong>Intl</strong> native object.</li>
</ul>

<p>This post requires some knowledge of new ES6 features such as array comprehensions,
arrow functions, spreads, etc. If you are new to ES6, I suggest brushing up on the
basics. I found Ariyaâs <a href="http://ariya.ofilabs.com/tag/es6">posts on ES6</a> to be useful.</p>

<h2 id="the-i18n-tag">The i18n tag</h2>

<p>Letâs start with an example of how we want to use the <code>i18n</code> tag.</p>

<div class="highlight"><pre><code class="javascript"><span class="kd">let</span> <span class="nx">name</span> <span class="o">=</span> <span class="s1">'Bob'</span><span class="p">;</span>
<span class="kd">let</span> <span class="nx">amount</span> <span class="o">=</span> <span class="mi">100</span><span class="p">;</span>
<span class="nx">i18n</span><span class="err">`</span><span class="nx">Hello</span> <span class="nx">$</span><span class="p">{</span><span class="nx">name</span><span class="p">},</span> <span class="nx">you</span> <span class="nx">have</span> <span class="nx">$</span><span class="p">{</span><span class="nx">amount</span><span class="p">}</span><span class="o">:</span><span class="nx">c</span><span class="p">(</span><span class="nx">CAD</span><span class="p">)</span> <span class="k">in</span> <span class="nx">your</span> <span class="nx">bank</span> <span class="nx">account</span><span class="p">.</span><span class="err">`</span>
</code></pre></div>

<p>Here, we are expecting to translate the literals into the language of our choosing.
We are also adding optional annotations to the substitutions. In this case, the
<code>amount</code> value has the <code>:c(CAD)</code> annotation, which marks it as a currency <strong>type</strong>
that is in Canadian Dollars (CAD).</p>

<p>If we translated this string to German, for example, we expect the result to be
<code>'Hallo Bob, Sie haben 1.234,56 $CA auf Ihrem Bankkonto.'</code> Note that the currency
amount has thousands and decimal separator as per German locale.</p>

<h2 id="implementation">Implementation</h2>

<p>Weâll allow configuration of our <code>i18n</code> tag through <code>I18n.use</code>.</p>

<div class="highlight"><pre><code class="javascript"><span class="kd">let</span> <span class="nx">messageBundle_de</span> <span class="o">=</span> <span class="p">{</span>
  <span class="s1">'Hello {0}, you have {1} in your bank account.'</span><span class="o">:</span> <span class="s1">'Hallo {0}, Sie haben {1} auf Ihrem Bankkonto.'</span>
<span class="p">};</span>

<span class="kd">let</span> <span class="nx">i18n</span> <span class="o">=</span> <span class="nx">I18n</span><span class="p">.</span><span class="nx">use</span><span class="p">({</span>
  <span class="nx">locale</span><span class="o">:</span> <span class="s1">'de-DE'</span><span class="p">,</span>
  <span class="nx">defaultCurrency</span><span class="o">:</span> <span class="s1">'EUR'</span><span class="p">,</span>
  <span class="nx">messageBundle</span><span class="o">:</span> <span class="nx">messageBundle_de</span><span class="p">});</span>

<span class="c1">// Now we can translate our string to German, with the Euro being the default currency</span>

<span class="nx">i18n</span><span class="err">`</span><span class="nx">Hello</span> <span class="nx">$</span><span class="p">{</span><span class="nx">name</span><span class="p">},</span> <span class="nx">you</span> <span class="nx">have</span> <span class="nx">$</span><span class="p">{</span><span class="nx">amount</span><span class="p">}</span><span class="o">:</span><span class="nx">c</span><span class="p">(</span><span class="nx">CAD</span><span class="p">)</span> <span class="k">in</span> <span class="nx">your</span> <span class="nx">bank</span> <span class="nx">account</span><span class="p">.</span><span class="err">`</span>
<span class="c1">// =&gt; Hallo Bob, Sie haben 1.234,56 $CA auf Ihrem Bankkonto.</span>
</code></pre></div>

<p>Letâs start with implementing the <code>I18n.use</code> method, which should return the
tag function.</p>

<div class="highlight"><pre><code class="javascript"><span class="kd">let</span> <span class="nx">I18n</span> <span class="o">=</span> <span class="p">{</span>
  <span class="nx">use</span><span class="p">({</span><span class="nx">lang</span><span class="p">,</span> <span class="nx">defaultCurrency</span><span class="p">,</span> <span class="nx">messageBundle</span><span class="p">})</span> <span class="p">{</span>
    <span class="nx">I18n</span><span class="p">.</span><span class="nx">lang</span> <span class="o">=</span> <span class="nx">lang</span><span class="p">;</span>
    <span class="nx">I18n</span><span class="p">.</span><span class="nx">defaultCurrency</span> <span class="o">=</span> <span class="nx">defaultCurrency</span><span class="p">;</span>
    <span class="nx">I18n</span><span class="p">.</span><span class="nx">messageBundle</span> <span class="o">=</span> <span class="nx">messageBundle</span><span class="p">;</span>
    <span class="k">return</span> <span class="nx">I18n</span><span class="p">.</span><span class="nx">translate</span><span class="p">;</span>
  <span class="p">},</span>

  <span class="nx">translate</span><span class="p">(</span><span class="nx">literals</span><span class="p">,</span> <span class="p">...</span><span class="nx">values</span><span class="p">)</span> <span class="p">{</span>
    <span class="kd">let</span> <span class="nx">translationKey</span> <span class="o">=</span> <span class="p">...;</span>
    <span class="kd">let</span> <span class="nx">translationString</span> <span class="o">=</span> <span class="nx">I18n</span><span class="p">.</span><span class="nx">messages</span><span class="p">[</span><span class="nx">translationKey</span><span class="p">];</span>

    <span class="k">if</span> <span class="p">(</span><span class="nx">translationString</span><span class="p">)</span> <span class="p">{</span>
      <span class="kd">let</span> <span class="nx">localizedValues</span> <span class="o">=</span> <span class="p">...;</span>
      <span class="k">return</span> <span class="nx">I18n</span><span class="p">.</span><span class="nx">_buildMessage</span><span class="p">(</span><span class="nx">translationString</span><span class="p">,</span> <span class="p">...</span><span class="nx">localizedValues</span><span class="p">);</span>
    <span class="p">}</span>

    <span class="k">return</span> <span class="s1">'Error: translation missing!'</span><span class="p">;</span>
  <span class="p">}</span>
<span class="p">};</span>
</code></pre></div>

<p>The <code>translationString</code> should match the keys in our message bundle. In our previous
German translation example, this would be <code>'Hello {0}, you have {1} in your bank account.'</code>.</p>

<p>The <code>localizedValues</code> array is a map of substitution <code>values</code> in their localized
forms. For example, in German <code>1000.12</code> should show as <code>'1.000,12'</code>.</p>

<p>Now, I will show a full implementation of the library.</p>

<div class="highlight"><pre><code class="javascript"><span class="c1">// Matches optional type annotations in i18n strings.</span>
<span class="c1">// e.g. i18n`This is a number ${x}:n(2)` formats x as number</span>
<span class="c1">//      with two fractional digits.</span>
<span class="kr">const</span> <span class="nx">typeInfoRegex</span> <span class="o">=</span> <span class="sr">/^:([a-z])(\((.+)\))?/</span><span class="p">;</span>

<span class="kd">let</span> <span class="nx">I18n</span> <span class="o">=</span> <span class="p">{</span>
  <span class="nx">use</span><span class="p">({</span><span class="nx">locale</span><span class="p">,</span> <span class="nx">defaultCurrency</span><span class="p">,</span> <span class="nx">messageBundle</span><span class="p">})</span> <span class="p">{</span>
    <span class="nx">I18n</span><span class="p">.</span><span class="nx">locale</span> <span class="o">=</span> <span class="nx">locale</span><span class="p">;</span>
    <span class="nx">I18n</span><span class="p">.</span><span class="nx">defaultCurrency</span> <span class="o">=</span> <span class="nx">defaultCurrency</span><span class="p">;</span>
    <span class="nx">I18n</span><span class="p">.</span><span class="nx">messageBundle</span> <span class="o">=</span> <span class="nx">messageBundle</span><span class="p">;</span>
    <span class="k">return</span> <span class="nx">I18n</span><span class="p">.</span><span class="nx">translate</span><span class="p">;</span>
  <span class="p">},</span>

  <span class="nx">translate</span><span class="p">(</span><span class="nx">literals</span><span class="p">,</span> <span class="p">...</span><span class="nx">values</span><span class="p">)</span> <span class="p">{</span>
    <span class="kd">let</span> <span class="nx">translationKey</span> <span class="o">=</span> <span class="nx">I18n</span><span class="p">.</span><span class="nx">_buildKey</span><span class="p">(</span><span class="nx">literals</span><span class="p">);</span>
    <span class="kd">let</span> <span class="nx">translationString</span> <span class="o">=</span> <span class="nx">I18n</span><span class="p">.</span><span class="nx">messageBundle</span><span class="p">[</span><span class="nx">translationKey</span><span class="p">];</span>

    <span class="k">if</span> <span class="p">(</span><span class="nx">translationString</span><span class="p">)</span> <span class="p">{</span>
      <span class="kd">let</span> <span class="nx">typeInfoForValues</span> <span class="o">=</span> <span class="nx">literals</span><span class="p">.</span><span class="nx">slice</span><span class="p">(</span><span class="mi">1</span><span class="p">).</span><span class="nx">map</span><span class="p">(</span><span class="nx">I18n</span><span class="p">.</span><span class="nx">_extractTypeInfo</span><span class="p">);</span>
      <span class="kd">let</span> <span class="nx">localizedValues</span> <span class="o">=</span> <span class="nx">values</span><span class="p">.</span><span class="nx">map</span><span class="p">((</span><span class="nx">v</span><span class="p">,</span> <span class="nx">i</span><span class="p">)</span> <span class="o">=&gt;</span> <span class="nx">I18n</span><span class="p">.</span><span class="nx">_localize</span><span class="p">(</span><span class="nx">v</span><span class="p">,</span> <span class="nx">typeInfoForValues</span><span class="p">[</span><span class="nx">i</span><span class="p">]));</span>
      <span class="k">return</span> <span class="nx">I18n</span><span class="p">.</span><span class="nx">_buildMessage</span><span class="p">(</span><span class="nx">translationString</span><span class="p">,</span> <span class="p">...</span><span class="nx">localizedValues</span><span class="p">);</span>
    <span class="p">}</span>

    <span class="k">return</span> <span class="s1">'Error: translation missing!'</span><span class="p">;</span>
  <span class="p">},</span>

  <span class="nx">_localizers</span><span class="o">:</span> <span class="p">{</span>
    <span class="nx">s</span> <span class="cm">/*string*/</span><span class="o">:</span> <span class="nx">v</span> <span class="o">=&gt;</span> <span class="nx">v</span><span class="p">.</span><span class="nx">toLocaleString</span><span class="p">(</span><span class="nx">I18n</span><span class="p">.</span><span class="nx">locale</span><span class="p">),</span>
    <span class="nx">c</span> <span class="cm">/*currency*/</span><span class="o">:</span> <span class="p">(</span><span class="nx">v</span><span class="p">,</span> <span class="nx">currency</span><span class="p">)</span> <span class="o">=&gt;</span> <span class="p">(</span>
      <span class="nx">v</span><span class="p">.</span><span class="nx">toLocaleString</span><span class="p">(</span><span class="nx">I18n</span><span class="p">.</span><span class="nx">locale</span><span class="p">,</span> <span class="p">{</span>
        <span class="nx">style</span><span class="o">:</span> <span class="s1">'currency'</span><span class="p">,</span>
        <span class="nx">currency</span><span class="o">:</span> <span class="nx">currency</span> <span class="o">||</span> <span class="nx">I18n</span><span class="p">.</span><span class="nx">defaultCurrency</span>
      <span class="p">})</span>
    <span class="p">),</span>
    <span class="nx">n</span> <span class="cm">/*number*/</span><span class="o">:</span> <span class="p">(</span><span class="nx">v</span><span class="p">,</span> <span class="nx">fractionalDigits</span><span class="p">)</span> <span class="o">=&gt;</span> <span class="p">(</span>
      <span class="nx">v</span><span class="p">.</span><span class="nx">toLocaleString</span><span class="p">(</span><span class="nx">I18n</span><span class="p">.</span><span class="nx">locale</span><span class="p">,</span> <span class="p">{</span>
        <span class="nx">minimumFractionDigits</span><span class="o">:</span> <span class="nx">fractionalDigits</span><span class="p">,</span>
        <span class="nx">maximumFractionDigits</span><span class="o">:</span> <span class="nx">fractionalDigits</span>
      <span class="p">})</span>
    <span class="p">)</span>
  <span class="p">},</span>

  <span class="nx">_extractTypeInfo</span><span class="p">(</span><span class="nx">literal</span><span class="p">)</span> <span class="p">{</span>
    <span class="kd">let</span> <span class="nx">match</span> <span class="o">=</span> <span class="nx">typeInfoRegex</span><span class="p">.</span><span class="nx">exec</span><span class="p">(</span><span class="nx">literal</span><span class="p">);</span>
    <span class="k">if</span> <span class="p">(</span><span class="nx">match</span><span class="p">)</span> <span class="p">{</span>
      <span class="k">return</span> <span class="p">{</span><span class="nx">type</span><span class="o">:</span> <span class="nx">match</span><span class="p">[</span><span class="mi">1</span><span class="p">],</span> <span class="nx">options</span><span class="o">:</span> <span class="nx">match</span><span class="p">[</span><span class="mi">3</span><span class="p">]};</span>
    <span class="p">}</span> <span class="k">else</span> <span class="p">{</span>
      <span class="k">return</span> <span class="p">{</span><span class="nx">type</span><span class="o">:</span> <span class="s1">'s'</span><span class="p">,</span> <span class="nx">options</span><span class="o">:</span> <span class="s1">''</span><span class="p">};</span>
    <span class="p">}</span>
  <span class="p">},</span>

  <span class="nx">_localize</span><span class="p">(</span><span class="nx">value</span><span class="p">,</span> <span class="p">{</span><span class="nx">type</span><span class="p">,</span> <span class="nx">options</span><span class="p">})</span> <span class="p">{</span>
    <span class="k">return</span> <span class="nx">I18n</span><span class="p">.</span><span class="nx">_localizers</span><span class="p">[</span><span class="nx">type</span><span class="p">](</span><span class="nx">value</span><span class="p">,</span> <span class="nx">options</span><span class="p">);</span>
  <span class="p">},</span>

  <span class="c1">// e.g. I18n._buildKey(['', ' has ', ':c in the']) == '{0} has {1} in the bank'</span>
  <span class="nx">_buildKey</span><span class="p">(</span><span class="nx">literals</span><span class="p">)</span> <span class="p">{</span>
    <span class="kd">let</span> <span class="nx">stripType</span> <span class="o">=</span> <span class="nx">s</span> <span class="o">=&gt;</span> <span class="nx">s</span><span class="p">.</span><span class="nx">replace</span><span class="p">(</span><span class="nx">typeInfoRegex</span><span class="p">,</span> <span class="s1">''</span><span class="p">);</span>
    <span class="kd">let</span> <span class="nx">lastPartialKey</span> <span class="o">=</span> <span class="nx">stripType</span><span class="p">(</span><span class="nx">literals</span><span class="p">[</span><span class="nx">literals</span><span class="p">.</span><span class="nx">length</span> <span class="o">-</span> <span class="mi">1</span><span class="p">]);</span>
    <span class="kd">let</span> <span class="nx">prependPartialKey</span> <span class="o">=</span> <span class="p">(</span><span class="nx">memo</span><span class="p">,</span> <span class="nx">curr</span><span class="p">,</span> <span class="nx">i</span><span class="p">)</span> <span class="o">=&gt;</span> <span class="err">`</span><span class="nx">$</span><span class="p">{</span><span class="nx">stripType</span><span class="p">(</span><span class="nx">curr</span><span class="p">)}{</span><span class="nx">$</span><span class="p">{</span><span class="nx">i</span><span class="p">}}</span><span class="nx">$</span><span class="p">{</span><span class="nx">memo</span><span class="p">}</span><span class="err">`</span><span class="p">;</span>

    <span class="k">return</span> <span class="nx">literals</span><span class="p">.</span><span class="nx">slice</span><span class="p">(</span><span class="mi">0</span><span class="p">,</span> <span class="o">-</span><span class="mi">1</span><span class="p">).</span><span class="nx">reduceRight</span><span class="p">(</span><span class="nx">prependPartialKey</span><span class="p">,</span> <span class="nx">lastPartialKey</span><span class="p">);</span>
  <span class="p">},</span>

  <span class="c1">// e.g. I18n._formatStrings('{0} {1}!', 'hello', 'world') == 'hello world!'</span>
  <span class="nx">_buildMessage</span><span class="p">(</span><span class="nx">str</span><span class="p">,</span> <span class="p">...</span><span class="nx">values</span><span class="p">)</span> <span class="p">{</span>
    <span class="k">return</span> <span class="nx">str</span><span class="p">.</span><span class="nx">replace</span><span class="p">(</span><span class="sr">/{(\d)}/g</span><span class="p">,</span> <span class="p">(</span><span class="nx">_</span><span class="p">,</span> <span class="nx">index</span><span class="p">)</span> <span class="o">=&gt;</span> <span class="nx">values</span><span class="p">[</span><span class="nb">Number</span><span class="p">(</span><span class="nx">index</span><span class="p">)]);</span>
  <span class="p">}</span>
<span class="p">};</span>

<span class="c1">// Usage</span>
<span class="kd">let</span> <span class="nx">messageBundle_fr</span> <span class="o">=</span> <span class="p">{</span>
  <span class="s1">'Hello {0}, you have {1} in your bank account.'</span><span class="o">:</span> <span class="s1">'Bonjour {0}, vous avez {1} dans votre compte bancaire.'</span>
<span class="p">};</span>

<span class="kd">let</span> <span class="nx">messageBundle_de</span> <span class="o">=</span> <span class="p">{</span>
<span class="s1">'Hello {0}, you have {1} in your bank account.'</span><span class="o">:</span> <span class="s1">'Hallo {0}, Sie haben {1} auf Ihrem Bankkonto.'</span>
<span class="p">};</span>

<span class="kd">let</span> <span class="nx">messageBundle_zh_Hant</span> <span class="o">=</span> <span class="p">{</span>
  <span class="s1">'Hello {0}, you have {1} in your bank account.'</span><span class="o">:</span> <span class="s1">'ä½ å¥½{0}ï¼ä½ æ{1}å¨æ¨çéè¡å¸³æ¶ã'</span>
<span class="p">};</span>

<span class="kd">let</span> <span class="nx">name</span> <span class="o">=</span> <span class="s1">'Bob'</span><span class="p">;</span>
<span class="kd">let</span> <span class="nx">amount</span> <span class="o">=</span> <span class="mf">1234.56</span><span class="p">;</span>
<span class="kd">let</span> <span class="nx">i18n</span><span class="p">;</span>

<span class="nx">i18n</span> <span class="o">=</span> <span class="nx">I18n</span><span class="p">.</span><span class="nx">use</span><span class="p">({</span><span class="nx">locale</span><span class="o">:</span> <span class="s1">'fr-CA'</span><span class="p">,</span> <span class="nx">defaultCurrency</span><span class="o">:</span> <span class="s1">'CAD'</span><span class="p">,</span> <span class="nx">messageBundle</span><span class="o">:</span> <span class="nx">messageBundle_fr</span><span class="p">});</span>
<span class="nx">console</span><span class="p">.</span><span class="nx">log</span><span class="p">(</span><span class="nx">i18n</span> <span class="err">`</span><span class="nx">Hello</span> <span class="nx">$</span><span class="p">{</span><span class="nx">name</span><span class="p">},</span> <span class="nx">you</span> <span class="nx">have</span> <span class="nx">$</span><span class="p">{</span><span class="nx">amount</span><span class="p">}</span><span class="o">:</span><span class="nx">c</span> <span class="k">in</span> <span class="nx">your</span> <span class="nx">bank</span> <span class="nx">account</span><span class="p">.</span><span class="err">`</span><span class="p">);</span>

<span class="nx">i18n</span> <span class="o">=</span> <span class="nx">I18n</span><span class="p">.</span><span class="nx">use</span><span class="p">({</span><span class="nx">locale</span><span class="o">:</span> <span class="s1">'de-DE'</span><span class="p">,</span> <span class="nx">defaultCurrency</span><span class="o">:</span> <span class="s1">'EUR'</span><span class="p">,</span> <span class="nx">messageBundle</span><span class="o">:</span> <span class="nx">messageBundle_de</span><span class="p">});</span>
<span class="nx">console</span><span class="p">.</span><span class="nx">log</span><span class="p">(</span><span class="nx">i18n</span> <span class="err">`</span><span class="nx">Hello</span> <span class="nx">$</span><span class="p">{</span><span class="nx">name</span><span class="p">},</span> <span class="nx">you</span> <span class="nx">have</span> <span class="nx">$</span><span class="p">{</span><span class="nx">amount</span><span class="p">}</span><span class="o">:</span><span class="nx">c</span> <span class="k">in</span> <span class="nx">your</span> <span class="nx">bank</span> <span class="nx">account</span><span class="p">.</span><span class="err">`</span><span class="p">);</span>

<span class="nx">i18n</span> <span class="o">=</span> <span class="nx">I18n</span><span class="p">.</span><span class="nx">use</span><span class="p">({</span><span class="nx">locale</span><span class="o">:</span> <span class="s1">'zh-Hant-CN'</span><span class="p">,</span> <span class="nx">defaultCurrency</span><span class="o">:</span> <span class="s1">'CNY'</span><span class="p">,</span> <span class="nx">messageBundle</span><span class="o">:</span> <span class="nx">messageBundle_zh_Hant</span><span class="p">});</span>
<span class="nx">console</span><span class="p">.</span><span class="nx">log</span><span class="p">(</span><span class="nx">i18n</span> <span class="err">`</span><span class="nx">Hello</span> <span class="nx">$</span><span class="p">{</span><span class="nx">name</span><span class="p">},</span> <span class="nx">you</span> <span class="nx">have</span> <span class="nx">$</span><span class="p">{</span><span class="nx">amount</span><span class="p">}</span><span class="o">:</span><span class="nx">c</span> <span class="k">in</span> <span class="nx">your</span> <span class="nx">bank</span> <span class="nx">account</span><span class="p">.</span><span class="err">`</span><span class="p">);</span>
</code></pre></div>

<p>One important thing to note here is that the <code>_localizers</code> are using the
native <a href="https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Global_Objects/Intl">Intl</a>
object to localize values.</p>

<h3 id="es-internationalization-api">ES Internationalization API</h3>

<p>The <a href="https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Global_Objects/Intl">Internationalization API</a>
adds new methods to native types, such as <code>Number.prototype.toLocaleString()</code> and <code>Date.prototype.toLocaleDateString()</code>.</p>

<p>These methods delegate to three objects: <code>Intl.Collator</code>, <code>Intl.NumberFormat</code>, and <code>Intl.DateTimeFormat</code>. The
collator object enables string comparisons to be language sensitive, while the other two are pretty self-explanatory.</p>

<p>An option of the <code>NumberFormat</code> we took advantage of in our implementation is the <code>style</code>. The style value can be
<code>decimal</code> (default), <code>currency</code>, or <code>percent</code>. In our case, we used both the currency and decimal support for the <code>:c</code> and
<code>:n</code> annotations respectively. For number localization, we also support rounding of fractional digits through the use of
<code>minimumFractionDigits</code> and <code>maximumFractionDigits</code> options.</p>

<h2 id="example-usages">Example usages</h2>

<p>Now, we can translate our original string into different languages by configuring
the <code>I18n</code> object using corresponding locales and message bundle.</p>

<div class="highlight"><pre><code class="javascript"><span class="kd">let</span> <span class="nx">name</span> <span class="o">=</span> <span class="s1">'Bob'</span><span class="p">;</span>
<span class="kd">let</span> <span class="nx">amount</span> <span class="o">=</span> <span class="mf">1234.56</span><span class="p">;</span>
<span class="kd">let</span> <span class="nx">i18n</span><span class="p">;</span>

<span class="kd">let</span> <span class="nx">messageBundle_fr</span> <span class="o">=</span> <span class="p">{</span>
  <span class="s1">'Hello {0}, you have {1} in your bank account.'</span><span class="o">:</span> <span class="s1">'Bonjour {0}, vous avez {1} dans votre compte bancaire.'</span>
<span class="p">};</span>
<span class="kd">let</span> <span class="nx">messageBundle_de</span> <span class="o">=</span> <span class="p">{</span>
  <span class="s1">'Hello {0}, you have {1} in your bank account.'</span><span class="o">:</span> <span class="s1">'Hallo {0}, Sie haben {1} auf Ihrem Bankkonto'</span>
<span class="p">};</span>
<span class="kd">let</span> <span class="nx">messageBundle_zh_Hant</span> <span class="o">=</span> <span class="p">{</span>
  <span class="s1">'Hello {0}, you have {1} in your bank account.'</span><span class="o">:</span> <span class="s1">'ä½ å¥½{0}ï¼ä½ æ{1}å¨æ¨çéè¡å¸³æ¶ã'</span>
<span class="p">};</span>

<span class="nx">i18n</span> <span class="o">=</span> <span class="nx">I18n</span><span class="p">.</span><span class="nx">use</span><span class="p">({</span><span class="nx">locale</span><span class="o">:</span> <span class="s1">'fr-CA'</span><span class="p">,</span> <span class="nx">defaultCurrency</span><span class="o">:</span> <span class="s1">'CAD'</span><span class="p">,</span> <span class="nx">messageBundle</span><span class="o">:</span> <span class="nx">messageBundle_fr</span><span class="p">});</span>
<span class="nx">i18n</span><span class="err">`</span><span class="nx">Hello</span> <span class="nx">$</span><span class="p">{</span><span class="nx">name</span><span class="p">},</span> <span class="nx">you</span> <span class="nx">have</span> <span class="nx">$</span><span class="p">{</span><span class="nx">amount</span><span class="p">}</span><span class="o">:</span><span class="nx">c</span> <span class="k">in</span> <span class="nx">your</span> <span class="nx">bank</span> <span class="nx">account</span><span class="p">.</span><span class="err">`</span><span class="p">;</span>
<span class="c1">// =&gt; 'Bonjour Bob, vous avez 1 234,56 $CA dans votre compte bancaire.'</span>

<span class="nx">i18n</span> <span class="o">=</span> <span class="nx">I18n</span><span class="p">.</span><span class="nx">use</span><span class="p">({</span><span class="nx">locale</span><span class="o">:</span> <span class="s1">'de-DE'</span><span class="p">,</span> <span class="nx">defaultCurrency</span><span class="o">:</span> <span class="s1">'EUR'</span><span class="p">,</span> <span class="nx">messageBundle</span><span class="o">:</span> <span class="nx">messageBundle_de</span><span class="p">});</span>
<span class="nx">i18n</span><span class="err">`</span><span class="nx">Hello</span> <span class="nx">$</span><span class="p">{</span><span class="nx">name</span><span class="p">},</span> <span class="nx">you</span> <span class="nx">have</span> <span class="nx">$</span><span class="p">{</span><span class="nx">amount</span><span class="p">}</span><span class="o">:</span><span class="nx">c</span> <span class="k">in</span> <span class="nx">your</span> <span class="nx">bank</span> <span class="nx">account</span><span class="p">.</span><span class="err">`</span><span class="p">;</span>
<span class="c1">// =&gt; 'Hallo Bob, Sie haben 1.234,56 â¬ auf Ihrem Bankkonto.'</span>

<span class="nx">i18n</span> <span class="o">=</span> <span class="nx">I18n</span><span class="p">.</span><span class="nx">use</span><span class="p">({</span><span class="nx">locale</span><span class="o">:</span> <span class="s1">'zh-Hant-CN'</span><span class="p">,</span> <span class="nx">defaultCurrency</span><span class="o">:</span> <span class="s1">'CNY'</span><span class="p">,</span> <span class="nx">messageBundle</span><span class="o">:</span> <span class="nx">messageBundle_zh_Hant</span><span class="p">});</span>
<span class="nx">i18n</span><span class="err">`</span><span class="nx">Hello</span> <span class="nx">$</span><span class="p">{</span><span class="nx">name</span><span class="p">},</span> <span class="nx">you</span> <span class="nx">have</span> <span class="nx">$</span><span class="p">{</span><span class="nx">amount</span><span class="p">}</span><span class="o">:</span><span class="nx">c</span> <span class="k">in</span> <span class="nx">your</span> <span class="nx">bank</span> <span class="nx">account</span><span class="p">.</span><span class="err">`</span><span class="p">;</span>
<span class="c1">// =&gt; 'ä½ å¥½Bobï¼ä½ æï¿¥1,234.56å¨æ¨çéè¡å¸³æ¶ã'</span>
</code></pre></div>

<p>Hereâs a link to an <a href="http://www.es6fiddle.net/i0rj2cc9/">ES6 Fiddle</a> for a live demo.</p>

<h2 id="concluding-thoughts">Concluding thoughts</h2>

<p>The tagged template string can be used to provide powerful DSL to JavaScript libraries.
I chose to try my hand at adding simple i18n support. There are still a lot of pieces
missing to provide a comprehensive i18n solution, but I hope to have shown how one
may go about building out this framework.</p>

<p>Pluralization can be a very tricky thing to tackle, and I am not qualified as of this
writing to understand all the nuances of implementing this feature. Perhaps with
a little more thinking, there may be some hope to add it in as well.</p>

<p>In any case, I do see a lot of potential in tagged template strings to add clean, clear
syntax for library functionalities.</p>

<p><small>Edit (2014/03/21): Fixed German translation and added a conclusion section.</small></p>

<p><small>Edit (2014/10/01): Updated examples to not use array comprehension since they were removed in ES6</small></p>
